#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://www.opensource.org/licenses/GNU General Public License v3 (GPLv3)-license
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>

from . import test_cli
from . import test_convert
from . import test_draw
from . import test_error
from . import test_projection
from . import test_projection_draw
from . import test_svg
from . import test_svgis

__all__ = [
    'test_cli',
    'test_convert',
    'test_draw',
    'test_error',
    'test_projection',
    'test_projection_draw',
    'test_svg',
    'test_svgis',
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015, Neil Freeman <contact@fakeisthenewreal.org>
import unittest
from svgis import clip
try:
    import shapely.geometry

    class ClipTestCase(unittest.TestCase):

        def setUp(self):
            self.bounds = (1, 1, 9, 9)
            self.coords = [[(2, 2), (100, 2), (11, 11), (12, 12), (2, 10), (2, 2)]]
            self.expected = [((2.0, 9.0), (9.0, 9.0), (9.0, 2.0), (2.0, 2.0), (2.0, 9.0))]
            self.gen = (x for x in self.coords[0])

        def testShapely(self):
            minx, miny, maxx, maxy = self.bounds
            bounds = {
                "type": "Polygon",
                "coordinates": [[(minx, miny), (minx, maxy), (maxx, maxy), (maxx, miny), (minx, miny)]]
            }
            coords = {
                "type": "Polygon",
                "coordinates": self.coords
            }

            b = shapely.geometry.shape(bounds)
            c = shapely.geometry.shape(coords)
            i = b.intersection(c)

            result = shapely.geometry.mapping(i)
            self.assertSequenceEqual(result['coordinates'], self.expected)

        def testClip(self):
            clipped = clip.clip({"type": "Polygon", "coordinates": self.coords}, self.bounds)
            self.assertSequenceEqual(clipped['coordinates'], self.expected)

        def testClipGeometry(self):
            geometry = {"type": "Polygon", "coordinates": self.coords}
            clipped = clip.clip(geometry, self.bounds)

            self.assertSequenceEqual(set(self.expected), set(clipped['coordinates']))

except ImportError:
    pass


class ExpandTestCase(unittest.TestCase):

    def setUp(self):
        self.bounds = (1, 1, 9, 9)
        self.coords = [[(2, 2), (100, 2), (11, 11), (12, 12), (2, 10), (2, 2)]]
        self.expected = [((2.0, 9.0), (9.0, 9.0), (9.0, 2.0), (2.0, 2.0), (2.0, 9.0))]
        self.gen = (x for x in self.coords[0])

    def testExpand(self):
        expanded = clip.expand(self.gen)

        try:
            expanded = expanded.tolist()
        except AttributeError:
            pass

        self.assertSequenceEqual([tuple(x) for x in expanded], self.coords[0])

    def testExpandGeometry(self):
        geom = {
            "type": "LineString",
            "coordinates": self.gen
        }

        expanded = clip.expand_geom(geom)

        try:
            expanded['coordinates'] = expanded['coordinates'].tolist()
        except AttributeError:
            pass

        self.assertSequenceEqual([tuple(x) for x in expanded['coordinates']], self.coords[0])

    def testExpandGeomCollection(self):
        gen = (x for x in self.coords[0])

        GC = {
            "type": "GeometryCollection",
            "geometries": [
                {
                    "type": "LineString",
                    "coordinates": self.gen
                },
                {
                    "type": "LineString",
                    "coordinates": gen
                }
            ]
        }
        a = clip.expand_geom(GC)

        assert len(a['geometries']) == 2

        try:
            b = a['geometries'][0]['coordinates'].tolist()

        except AttributeError:
            b = a['geometries'][0]['coordinates']

        self.assertSequenceEqual(b[0], [2, 2])

    def testStrangeGeometry(self):
        with self.assertRaises(NotImplementedError):
            clip.expand_geom({"type": "Magic", "coordinates": [], "geometries": []})


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2015-16, Neil Freeman <contact@fakeisthenewreal.org>

from functools import partial

try:
    import visvalingamwyatt as vw
except ImportError:
    pass
from . import utils


def updatebounds(old, new):
    '''
    Extend old with any more distant values from newpoints.
    Also replace any missing min/max points in old with values from new.
    '''
    bounds = []
    inf = float('inf')
    neginf = inf * -1

    # python3 gives TypeError when using None in min/max
    # This contraption avoids that problem.
    # List comp below replaces Nones in bounds with real values in new or old
    for n, m in zip(new[:2], old[:2]):
        try:
            if neginf in (m, n):
                bounds.append(max(n, m))
                continue

            bounds.append(min(n, m))
        except TypeError:
            bounds.append(None)

    for n, m in zip(new[2:], old[2:]):
        try:
            if inf in (m, n):
                bounds.append(min(n, m))
                continue

            bounds.append(max(n, m))
        except TypeError:
            bounds.append(None)

    if any(not v for v in bounds):
        bounds = list((a or b or c) for a, b, c in zip(bounds, new, old))

    return bounds


def extend_bbox(bbox, ext=100):
    '''
    Widen the bounding box just a little bit. Assumes the bbox is in feet or meters or something.
    '''
    return bbox[0] - ext, bbox[1] - ext, bbox[2] + ext, bbox[3] + ext


def bounds_to_ring(minx, miny, maxx, maxy):
    """Convert min, max points to a boundary ring."""

    xs, ys = list(utils.between(minx, maxx)), list(utils.between(miny, maxy))

    left_top = [(minx, y) for y in ys] + [(x, maxy) for x in xs][1:]

    ys.reverse()
    xs.reverse()

    return left_top + [(maxx, y) for y in ys] + [(x, miny) for x in xs]


def simplifier(ratio):
    '''
    Create a simplification function, if visvalingamwyatt is available.
    Otherwise, return a noop function.

    Args:
        ratio (int): Between 1 and 99
    '''
    try:
        # put this first to get NameError out of the way
        simplify = vw.simplify_geometry

        if ratio is None or ratio >= 100 or ratio < 1:
            raise ValueError

        return partial(simplify, ratio=ratio/100.)

    except (TypeError, ValueError, NameError):
        return None


def bbox_covers(b1, b2):
    '''
    Return True if b1 covers b2.

    Args:
        b1 (tuple): A bounding box (minx, miny, maxx, maxy)
        b2 (tuple): A bounding box
    '''
    return b1[0] <= b2[0] and b1[1] <= b2[1] and b1[2] >= b2[2] and b1[3] >= b2[3]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

import re
from string import ascii_letters
import os.path
import logging
try:
    import xml.etree.cElementTree as ElementTree
except ImportError:
    import xml.etree.ElementTree as ElementTree
import tinycss
from . import dom

'''
Utilities for messing with SVG styling.
'''


def _register():
    ElementTree.register_namespace('', dom.SVG_NS)


def sanitize(string):
    '''
    Make input safe of use in an svg ID or class field.
    Replaces blocks of whitespace with an underscore (``_``).
    If the first character isn't an ascii letter, dash (``-``)
    or underscore (``_``), an underscore is added to the beginning.

    Args:
        string (mixed): Input to sanitize. Will be immediately coerced to
                        unicode. If that fails, returns an empty string.
    '''
    try:
        string = re.sub(r'\s+', u'_', unicode(string))
        string = re.sub(r'(\.|#|")', '', string)
        string = string if string[0] in ('_-' + ascii_letters) else '_' + string
        return unicode(string).strip()

    except (AttributeError, IndexError):
        return u''


def construct_classes(classes, properties):
    '''
    Build a class string for an element using the properties. Class names
    take the form CLASS_PROPERTY. If a given class isn't found in properties,
    the class name is added (e.g. CLASS).

    Args:
        classes (Sequence): Column names to include in the class list
        properties (dict): A single feature's properties.

    Returns:
        str (unicode in Python 2)
    '''
    f = u'{}_{}'
    classed = (f.format(p, properties[p]) if p in properties else p for p in classes)
    sanitized = (sanitize(w) for w in classed)
    return u' '.join(sanitized)


def pick(style):
    '''
    Fetch a CSS string.

    Args:
        style (str): Either a CSS string or the path of a css file.
    '''
    try:
        _, ext = os.path.splitext(style)
        if ext == '.css':
            with open(style) as f:
                return f.read()

    except AttributeError:
        # Probably style is None.
        return None

    except IOError:
        logging.getLogger('svgis').warning("Couldn't read %s, proceeding with default style", style)

    return style


def rescale(svgfile, factor):
    _register()
    try:
        svg = ElementTree.parse(svgfile)
        scalar = 'scale({})'.format(factor)
        g = svg.getroot().find(dom.ns('g'))
        g.attrib['transform'] = (g.attrib.get('transform') + ' ' + scalar).strip()

    except AttributeError:
        raise

    return ElementTree.tostring(svg.getroot(), encoding='utf-8').decode('utf-8')


def add_style(svgfile, style, replace=False):
    '''
    Add to or replace the CSS style in an SVG file.

    Args:
        svgfile (str): Path to an SVG file or an SVG string.
        newstyle (str): CSS string, or path to CSS file.
        replace (bool): If true, replace the existing CSS with newstyle (default: False)
    '''
    _register()

    if style == '-':
        style = '/dev/stdin'

    root, ext = os.path.splitext(style)

    if ext == '.css' or root == '/dev/stdin':
        with open(style) as f:
            style = f.read()

    try:
        svg = ElementTree.parse(svgfile).getroot()
    except IOError:
        try:
            svg = ElementTree.fromstring(svgfile)
        except UnicodeDecodeError:
            svg = ElementTree.fromstring(svgfile.encode('utf-8'))

    if svg.find(dom.ns('defs')) is not None:
        defs = svg.find(dom.ns('defs'))

    else:
        defs = ElementTree.Element(dom.ns('defs'))
        svg.insert(0, defs)

    if defs.find(dom.ns('style')) is not None:
        style_element = defs.find(dom.ns('style'))

        if not replace:
            # Append CSS.
            existing = style_element.text or ''
            style = dom.cdata(existing + ' ' + style)
            style_element.text = ''

    else:
        style_element = ElementTree.Element(dom.ns('style'))
        defs.append(style_element)

    style_element.text = dom.cdata(style)

    return ElementTree.tostring(svg, encoding='utf-8').decode('utf-8')


def inline(svg, style=None):
    '''
    Inline the CSS rules in an SVG. This is a very rough operation,
    and full css precedence rules won't be respected. Ignores sibling
    operators (``~``, ``+``), psuedo-selectors (e.g. ``:first-child``), and
    attribute selectors (e.g. ``.foo[name=bar]``). Works best with rules like:

    * ``.class``
    * ``tag``
    * ``tag.class``
    * ``#layer .class``
    * ``#layer tag``

    Args:
        svg (string): An SVG document.
        css (string): CSS to use, instead of the CSS in the <defs> element of the SVG.
    '''
    _register()
    try:
        doc = ElementTree.fromstring(svg.encode('utf-8'))
        wrap = doc.find('./' + dom.ns('g'))

        if not style:
            path = './' + dom.ns('defs') + '/' + dom.ns('style')
            try:
                style = doc.findall(path).pop().text
            except IndexError:
                style = ''

        stylesheet = _parse_css(style)

        for rule in stylesheet.rules:
            dom.apply_rule(wrap, rule)

        return ElementTree.tostring(doc, encoding='utf-8').decode('utf-8')

    # Return plain old SVG.
    except (AttributeError, NameError) as e:
        logging.getLogger('svgis').warning("Unable to inline CSS becuase: %s", e)
        return svg


def _parse_css(stylesheet):
    mini = re.sub(r'\s+([,>~+])\s+', r'\1', stylesheet)
    return tinycss.make_parser().parse_stylesheet(mini)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://opensource.org/licenses/GPL-3.0
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

'''
Create string versions of SVG elements.
'''


def _element(tag, contents=None, **kwargs):
    '''
    Draw an element, optionally wrapping contents.

    Args:
        tag (str): tag name
        contents (str): contents to be wrapped in the tag
        kwargs: to be transformed into attributes

    Returns:
        str (unicode in Python 2)
    '''
    attribs = toattribs(**kwargs)
    if contents:
        return u'<{0}{1}>{2}</{0}>'.format(tag, attribs, contents)
    else:
        return u'<{0}{1}/>'.format(tag, attribs)


def _round(i, precision):
    if precision is None:
        return i
    else:
        return round(i, precision)


def _fmtcoord(precision):
    if precision is None:
        return u'{0[0]},{0[1]}'
    else:
        return u'{{0[0]:.{0}f}},{{0[1]:.{0}f}}'.format(precision)


def circle(point, precision=None, **kwargs):
    '''
    Create a svg circle element. Keyword arguments are mapped to attributes.

    Args:
        point (tuple): The center of the circle
        precision (int): rounding precision

    Returns:
        str
    '''
    return _element(u'circle', cx=_round(point[0], precision), cy=_round(point[1], precision), **kwargs)


def text(string, start, precision=None, **kwargs):
    '''
    Create an svg text element.

    Args:
        string (str): text for element
        start (tuple): starting coordinate

    Returns:
        str
    '''
    start = [_round(i, precision) for i in start]
    return _element(u'text', string, x=start[0], y=start[1], **kwargs)


def rect(start, width, height, precision=None, **kwargs):
    '''
    Create an svg rect element.

    Args:
        start (tuple): starting coordinate
        width (int): rect width
        height (int): rect height
        precision (int): rounding precision

    Returns:
        str
    '''
    start = [_round(i, precision) for i in start]
    width = _round(width, precision)
    height = _round(height, precision)

    return _element(u'rect', x=start[0], y=start[1], width=width, height=height, **kwargs)


def line(start, end, precision=None, **kwargs):
    '''
    Create an svg line element.

    Args:
        start (tuple): starting coordinate
        end (tuple): ending coordinate
        precision (int): rounding precision

    Returns:
        str
    '''
    start = [_round(i, precision) for i in start]
    end = [_round(i, precision) for i in end]

    return _element(u'line', x1=start[0], y1=start[1], x2=end[0], y2=end[1], **kwargs)


def _isstr(x):
    return isinstance(x, basestring)


def path(coordinates, precision=None, **kwargs):
    '''
    Create an svg path element as a string.

    Args:
        coordinates (Sequence): A sequence of coordinates and string instructions
        precision (int): rounding precision

    Returns:
        str
    '''
    fmt = _fmtcoord(precision)
    coords = (i if _isstr(i) else fmt.format(i) for i in coordinates)
    return _element(u'path', d=' '.join(coords), **kwargs)


def polyline(coordinates, precision=None, **kwargs):
    '''
    Create an svg polyline element

    Args:
        coordinates (Sequence): x, y coordinates
        precision (int): rounding precision

    Returns:
        str
    '''
    fmt = _fmtcoord(precision)
    points = u' '.join(fmt.format(c) for c in coordinates)
    return _element(u'polyline', points=points, **kwargs)


def polygon(coordinates, precision=None, **kwargs):
    '''
    Create an svg polygon element

    Args:
        coordinates (Sequence): x, y coordinates
        precision (int): rounding precision

    Returns:
        str
    '''
    fmt = _fmtcoord(precision)
    points = u' '.join(fmt.format(c) for c in coordinates)
    return _element(u'polygon', points=points, **kwargs)


def toattribs(**kwargs):
    attribs = u' '.join(u'{}="{}"'.format(k, v) for k, v in kwargs.items() if v is not None)

    if len(attribs) > 0:
        return ' ' + attribs
    else:
        return attribs


def defstyle(style=None):
    '''
    Create a defs element that wraps a CSS style.

    Args:
        style (string): A CSS string.

    Returns:
        unicode
    '''
    if style:
        return u'<defs><style type="text/css"><![CDATA[{}]]></style></defs>'.format(style)
    else:
        return u'<defs />'


def group(members=None, **kwargs):
    '''
    Create a group with the given scale and translation.

    Args:
        members (Sequence): unicode SVG elements
        kwargs (dict): elements of this dictionary will be converted to
                        attributes of the group, i.e. key="value".

    Returns:
        unicode
    '''
    members = members or ''
    return _element(u'g', u''.join(members), **kwargs)


def drawing(size, members, precision=None, viewbox=None, style=None):
    '''
    Create an SVG element.

    Args:
        size (tuple): width, height
        members (list): Strings to add to output.
        viewbox (Sequence): Four coordinates that describe an SVG viewBox.
        style (string): CSS string.
    '''
    kwargs = {
        'width': size[0],
        'height': size[1],
        'baseProfile': 'full',
        'version': '1.1',
        'xmlns': 'http://www.w3.org/2000/svg',
    }

    if precision:
        fmt = ('{:.{precision}f}',)
    else:
        fmt = ('{:f}',)

    if viewbox:
        kwargs['viewBox'] = (','.join(fmt * 4)).format(*viewbox, precision=precision)

    contents = defstyle(style) + u''.join(members)
    return _element(u'svg', contents, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# This file is part of svgis.
# https://github.com/fitnr/svgis

# Licensed under the GNU General Public License v3 (GPLv3) license:
# http://www.opensource.org/licenses/GNU General Public License v3 (GPLv3)-license
# Copyright (c) 2016, Neil Freeman <contact@fakeisthenewreal.org>

from __future__ import division
import json
from functools import partial
import fiona.transform
from . import projection, utils


"""
Draw regular-interval graticules for a coordinate system
"""


def graticule(bounds, step, crs=None):
    '''
    Draw graticules.

    Args:
        bounds (tuple): In WGS84 coordinates.
        step (int): Distance between graticule lines, in the output projection.
        crs (str): A projection specification.

    Returns:
        A generator that yields GeoJSON-like dicts of graticule features.
    '''
    if crs == 'file':
        raise ValueError("'file' is not a valid option for projecting graticules.")

    if crs:
        method, out_crs = projection.pick(crs)
        if method in ('local', 'utm'):
            out_crs = projection.generatecrs(*bounds, proj_method=method)

        bounds = projection.transform_bounds({'init': u'epsg:4326'}, out_crs, bounds)
        unproject = partial(fiona.transform.transform, out_crs, {'init': u'epsg:4326'})

    else:
        def unproject(x, y):
            return x, y

    minx, miny, maxx, maxy = bounds

    minx, miny = utils.modfloor(minx, step), utils.modfloor(miny, step)
    maxx, maxy = utils.modceil(maxx, step), utils.modceil(maxy, step)

    frange = partial(utils.frange, cover=True)

    for i, X in enumerate(frange(minx, maxx + step, step), 1):
        coords = unproject(*zip(*[(X, y) for y in frange(miny, maxy + step, step/2)]))
        yield _feature(i, zip(*coords), axis='x', coord=X)

    for i, Y in enumerate(frange(miny, maxy + step, step), i+1):
        coords = unproject(*zip(*[(x, Y) for x in frange(minx, maxx + step, step/2)]))
        yield _feature(i, zip(*coords), axis='y', coord=Y)


def _feature(i, coords, axis=None, coord=None):
    return {
        'geometry': {
            'type': 'LineString',
            'coordinates': list(coords)
        },
        'properties': {
            'axis': axis,
            'coord': coord
        },
        'type': 'Feature',
        'id': i
    }


def layer(bounds, step, crs=None):
    return {
        "type": "FeatureCollection",
        "crs": {
            "type": "name",
            "properties": {
                "name": "urn:ogc:def:crs:OGC:1.3:CRS84"
            },
        },
        "features": list(graticule(bounds, step, crs))
    }


def geojson(bounds, step, crs=None):
    '''
    Create a geojson with graticules of the given bounds.

    Returns:
        str
    '''
    return json.dumps(layer(bounds, step, crs))
